local powerTypes = {
  [POWERTYPE_HEALTH]  = {friendly = "health",  label = "Health"},
  [POWERTYPE_MAGICKA] = {friendly = "magicka", label = "Magicka"},
  [POWERTYPE_STAMINA] = {friendly = "stamina", label = "Stamina"}
}

GGF.Unit = ZO_Object:Subclass()

-- Create and Initialize a NEW Unit Object (OOP FTW)
function GGF.Unit:New(unitName, baseTemplate, parent)
  local self = ZO_Object.New( self )
  self.unitName = unitName
  self.parent = parent
  self.frames = {}
  self:Controls()
  return self
end


----------------------------------------
-- Helper Functions
----------------------------------------

local function RemoveBuff(buff)
  buff:SetHidden(true)
end
local function CreateBuff(pool)
  local unit = pool.parent
  local id = pool:GetNextControlId()
  local buff = GGF.Window:CreateBackdrop("GGF_"..unit.unitName.."Buff"..id, unit.frames.main, unit.template.Buff.Buff); 
  buff.back  = GGF.Window:CreateTexture("GGF_"..unit.unitName.."BuffBk"..id, buff, unit.template.Buff.Buff.Back)
  buff.icon  = GGF.Window:CreateTexture("GGF_"..unit.unitName.."BuffTx"..id, buff, unit.template.Buff.Buff.Icon)
  buff.text  = GGF.Window:CreateLabel(  "GGF_"..unit.unitName.."BuffLb"..id, buff, unit.template.Buff.Buff.Text)
  buff.id = id

  -- Anchor this buff
  local padding = 3
  buff:ClearAnchors()
  local orderX = GGF.SavedVars[unit.unitName..'_Buff_OrderX'] == GGF.Locale['OrderX'][1] and {'LEFT','RIGHT'} or {'RIGHT','LEFT'}
  local orderY = GGF.SavedVars[unit.unitName..'_Buff_OrderY'] == GGF.Locale['OrderY'][1] and {'TOP','BOTTOM'} or {'BOTTOM','TOP'}
  local wide = math.floor((unit.template.Buff.BuffContainer.Width+padding) / (unit.template.Buff.Buff.Width+padding))
  local myAnchor = _G[orderY[1]..orderX[1]]
  if id == 1 then
    buff:SetAnchor(myAnchor, unit.frames.buffContainer, _G[orderY[1]..orderX[1]], 0, 0)
  elseif id % wide ~= 1 then
    local p = GGF.SavedVars[unit.unitName..'_Buff_OrderX'] == GGF.Locale['OrderX'][1] and padding or (-1*padding)
    buff:SetAnchor(myAnchor, pool:AcquireObject(buff.id-1), _G[orderY[1]..orderX[2]], p, 0)
  else
    local p = GGF.SavedVars[unit.unitName..'_Buff_OrderX'] == GGF.Locale['OrderX'][1] and (-1*padding) or padding
    buff:SetAnchor(myAnchor, pool:AcquireObject(buff.id-wide), _G[orderY[2]..orderX[1]], 0, p)
  end

  buff.SecondsToString = function(seconds)
    if seconds > 60 then
      return string.format("%d:%.2d", math.floor(seconds/60), math.floor(seconds%60))
    end
    return string.format(":%.2d", math.floor(seconds%60))
  end

  -- Setup Handler(s)
  buff.lastUpdate = 0
  buff:SetHandler("OnUpdate", function(self, updateTime)
    if (updateTime - self.lastUpdate) >= .5 and not buff.neverEnding then -- Buffering update timing
      local timeLeft = self.SecondsToString(self.timeEnding - updateTime)
      self.text:SetText(timeLeft)
      self.lastUpdate = updateTime
    end
  end)
  buff:SetMouseEnabled(true)
  buff:SetHandler("OnMouseEnter", ZO_Options_OnMouseEnter)
  buff:SetHandler("OnMouseExit",  ZO_Options_OnMouseExit)
  return buff
end

----------------------------------------
-- Functions
----------------------------------------

-- Load Unit Data
function GGF.Unit:Load(unitTag, graceful)
  if graceful == true and self.name == GetUnitName(unitTag) then return end
  
  self:ClearBuffs() -- Clear any residual buffs (if any)

  self.unitTag = unitTag
  
  self:SetName( GetUnitName(self.unitTag) )
  self:SetConnected( IsUnitOnline(self.unitTag) )
  self:SetDeath( IsUnitDead(self.unitTag) )
  self:SetLevel( GetUnitLevel(self.unitTag), GetUnitVeteranRank(self.unitTag) )             -- Level / Experience
  self:SetRange( IsUnitInGroupSupportRange(self.unitTag) )                                  -- Is Within Support Range
  self:SetStealth( GetUnitStealthState(self.unitTag) ~= STEALTH_STATE_NONE )

  if self.unitName == "Target" then
    self:SetCaption( GetUnitCaption(self.unitTag) or GetUnitTitle(self.unitTag) )
    self:SetAllianceRank( GetUnitAvARank(self.unitTag) )
    self:SetDifficulty( GetUnitDifficulty(self.unitTag) )
  end
  
  if IsUnitPlayer(self.unitTag) then
    self:SetClass( GetUnitClassId(self.unitTag) )                                           -- Class Texture
    self:SetLeader( IsUnitGroupLeader(self.unitTag) )                                       -- Is Group Leader (Non-Players are never Group Leaders)
  else 
    self:SetClass( nil )
    self:SetLeader( false )
  end

  self:SetPower( nil, POWERTYPE_HEALTH, GetUnitPower(self.unitTag, POWERTYPE_HEALTH) )      -- Set Health
  self:UpdateShield( GetUnitAttributeVisualizerEffectInfo(self.unitTag, ATTRIBUTE_VISUAL_POWER_SHIELDING, STAT_MITIGATION, ATTRIBUTE_HEALTH, POWERTYPE_HEALTH) )
  self:UpdateRegen( STAT_HEALTH_REGEN_COMBAT,  ATTRIBUTE_HEALTH,  POWERTYPE_HEALTH )
  self:UpdateRegen( STAT_MAGICKA_REGEN_COMBAT, ATTRIBUTE_MAGICKA, POWERTYPE_MAGICKA )
  self:UpdateRegen( STAT_STAMINA_REGEN_COMBAT, ATTRIBUTE_STAMINA, POWERTYPE_STAMINA )
  if self.unitName == "Player" then
    self:SetPower( nil, POWERTYPE_MAGICKA, GetUnitPower(self.unitTag, POWERTYPE_MAGICKA) )  -- Set Magicka
    self:SetPower( nil, POWERTYPE_STAMINA, GetUnitPower(self.unitTag, POWERTYPE_STAMINA) )  -- Set Stamina
    
    if IsUnitVeteran(self.unitTag) then
      self:SetExp( GetUnitVeteranPoints(self.unitTag), GetUnitVeteranPointsMax(self.unitTag), true )
    else
      self:SetExp( GetUnitXP(self.unitTag), GetUnitXPMax(self.unitTag), false )
    end
    self:SetMounted( IsMounted() )
    self:SetMountPower( nil, POWERTYPE_MOUNT_STAMINA, GetUnitPower(self.unitTag, POWERTYPE_MOUNT_STAMINA) )      -- Set Mount Stamina
  end
  
  if GGF.SavedVars['Display_Buffs'] and (self.unitName == "Player" or self.unitName == "Target") then
    self.buffPool = ZO_ObjectPool:New(CreateBuff, RemoveBuff)
    self.buffPool.parent = self
    self.debuffPool = ZO_ObjectPool:New(CreateBuff, RemoveBuff)
    self.debuffPool.parent = self
    self:UpdateBuffs(unitTag)
  end

  self.frames.main:SetHidden(false)
end

function GGF.Unit:Reload()
  self:Load(self.unitTag)
end

function GGF.Unit:Unload()
  self.name = ""
  self.frames.main:SetHidden(true)
  self:ClearBuffs()
end

function GGF.Unit:ClearBuffs()
  if self.template.Buff ~= false and self.buffPool and self.debuffPool then 
    self.buffPool:ReleaseAllObjects()
    self.debuffPool:ReleaseAllObjects()
  end
end

--------------------
-- Setters & Getters
--------------------

function GGF.Unit:SetPower( powerIndex, powerType,  powerValue, powerMax, powerEffectiveMax )
  if not self.template[powerTypes[powerType].label] then return end
  self[powerTypes[powerType].friendly] = {current = powerValue, max = powerEffectiveMax, percent = ( ( powerValue / powerEffectiveMax ) * 100 )}
  
  self.frames[powerTypes[powerType].friendly.."St"]:SetWidth( ( self[powerTypes[powerType].friendly].percent / 100 ) * self.template[powerTypes[powerType].label].Bar.Width )
  
  if self.frames[powerTypes[powerType].friendly.."LbOne"] then 
    self.frames[powerTypes[powerType].friendly.."LbOne"]:SetText( self:FormatStatusBarLabel(GGF.SavedVars['Bar_Format'], powerTypes[powerType].friendly) ) 
  end
  if self.frames[powerTypes[powerType].friendly.."LbTwo"] then 
    self.frames[powerTypes[powerType].friendly.."LbTwo"]:SetText( self:FormatStatusBarLabel(GGF.SavedVars['Bar_Format_Two'], powerTypes[powerType].friendly) )
  end
end

function GGF.Unit:SetMountPower( powerIndex, powerType, powerValue, powerMax, powerEffectiveMax )
  if not self.template.Mount then return end
  self.mount = {current = powerValue, max = powerEffectiveMax, percent = ( ( powerValue / powerEffectiveMax ) * 100 )}
  local barWidth = GGF.SavedVars['Player_Mount_Seperated'] and self.template.Mount.BarArea.Bar.Width or self.template.Mount.Bar.Width
  self.frames.mountSt:SetWidth( ( self.mount.percent / 100 ) * barWidth )
end

function GGF.Unit:SetMounted( isMounted )
  if not self.template.Mount then return end
  self.isMounted = isMounted
  self.frames.mount:SetHidden( not self.isMounted )
  if self.template.Experience ~= false and not GGF.SavedVars['Player_Mount_Seperated'] then
    GGF.Window:SetOffsetY(self.frames.experienceBd, isMounted and self.template.Experience.OffsetY or self.template.Experience.UnmountedOffsetY )
  end
end

function GGF.Unit:UpdateShield( value, maxValue )
  if not self.template.Health.Shield then return end
  if value == nil then value = 0 end
  if maxValue == nil then maxValue = 0 end
  self.shield = {["current"] = value, ["max"] = maxValue, ["percent"] = math.floor( ( value / maxValue ) * 100 )}
  self.frames.shield:SetWidth( math.min( (value/self.health.max ) , 1 ) * self.template.Health.Bar.Width )
  if self.frames['healthLbOne'] then self.frames['healthLbOne']:SetText( self:FormatStatusBarLabel(GGF.SavedVars['Bar_Format'], 'health') ) end
  if self.frames['healthLbTwo'] then self.frames['healthLbTwo']:SetText( self:FormatStatusBarLabel(GGF.SavedVars['Bar_Format_Two'], 'health') ) end
end

function GGF.Unit:UpdateRegen( statType, attributeType, powerType )
  if GGF.SavedVars['Display_Power_Regens'] == false then return end
  if not self.template[powerTypes[powerType].label] or not self.template[powerTypes[powerType].label]["Regen"] or not self.template[powerTypes[powerType].label]["Degen"] then return end
  
  value = (GetUnitAttributeVisualizerEffectInfo(self.unitTag, ATTRIBUTE_VISUAL_INCREASED_REGEN_POWER, statType, attributeType, powerType) or 0)
        + (GetUnitAttributeVisualizerEffectInfo(self.unitTag, ATTRIBUTE_VISUAL_DECREASED_REGEN_POWER, statType, attributeType, powerType) or 0)
  
  if value > 0 then
    self.frames[powerTypes[powerType].friendly.."RegenTexture"]:SetHidden(false)
    self.frames[powerTypes[powerType].friendly.."RegenTime"]:PlayFromStart()
    self.frames[powerTypes[powerType].friendly.."DegenTexture"]:SetHidden(true)
    self.frames[powerTypes[powerType].friendly.."DegenTime"]:Stop()
  elseif value < 0 then
    self.frames[powerTypes[powerType].friendly.."RegenTexture"]:SetHidden(true)
    self.frames[powerTypes[powerType].friendly.."RegenTime"]:Stop()
    self.frames[powerTypes[powerType].friendly.."DegenTexture"]:SetHidden(false)
    self.frames[powerTypes[powerType].friendly.."DegenTime"]:PlayFromStart()
  else
    self.frames[powerTypes[powerType].friendly.."RegenTexture"]:SetHidden(true)
    self.frames[powerTypes[powerType].friendly.."RegenTime"]:Stop()
    self.frames[powerTypes[powerType].friendly.."DegenTexture"]:SetHidden(true)
    self.frames[powerTypes[powerType].friendly.."DegenTime"]:Stop()
  end
end

function GGF.Unit:SetName( name )
  if not self.template.Name then return end

  self.name = name
  GGF.Window:SetLabelText( self.frames.nameLb, self.name, (string.sub(self.unitName,0,10) ~= "LargeGroup") )

  self.reactionType = GetUnitReaction( self.unitTag )
  if( self.unitName == "Target" and (self.reactionType == UNIT_REACTION_FRIENDLY or self.reactionType == UNIT_REACTION_NPC_ALLY) )then
    self.frames.nameLb:SetColor(GGF.SavedVars['Target_FontColor_Friendly'][1], GGF.SavedVars['Target_FontColor_Friendly'][2], GGF.SavedVars['Target_FontColor_Friendly'][3], GGF.SavedVars['Target_FontColor_Friendly'][4])
  elseif self.unitName == "Target" and self.reactionType == UNIT_REACTION_HOSTILE then
    self.frames.nameLb:SetColor(GGF.SavedVars['Target_FontColor_Hostile'][1], GGF.SavedVars['Target_FontColor_Hostile'][2], GGF.SavedVars['Target_FontColor_Hostile'][3], GGF.SavedVars['Target_FontColor_Hostile'][4])
  elseif self.unitName == "Target" and self.reactionType == UNIT_REACTION_INTERACT then
    self.frames.nameLb:SetColor(GGF.SavedVars['Target_FontColor_Interact'][1], GGF.SavedVars['Target_FontColor_Interact'][2], GGF.SavedVars['Target_FontColor_Interact'][3], GGF.SavedVars['Target_FontColor_Interact'][4])
  else
    self.frames.nameLb:SetColor(GGF.SavedVars['Target_FontColor'][1], GGF.SavedVars['Target_FontColor'][2], GGF.SavedVars['Target_FontColor'][3], GGF.SavedVars['Target_FontColor'][4])
  end
end

function GGF.Unit:SetLevel( level, rank )
  if not self.isOnline or not self.template.Level then return end
  self.level = level
  self.vlevel = rank
  if self.vlevel > 0 then
    local size = self.template.Level.FontSize + 6
    self.frames.levelLb:SetText( string.format("|t%d:%d:%s|t%s", size, size, 'ggFrames\\theme\\textures\\LevelVeteran.dds', self.vlevel) )
  else
    local size = self.template.Level.FontSize + 2
    self.frames.levelLb:SetText( string.format("|t%d:%d:%s|t%s", size, size, 'ggFrames\\theme\\textures\\LevelNormal.dds', self.level) )
  end
  self.frames.levelLb:SetHidden( self.reactionType == UNIT_REACTION_FRIENDLY or self.reactionType == UNIT_REACTION_NPC_ALLY )
end

function GGF.Unit:SetCaption( caption )
  if not self.template.Caption then return end
  self.caption = caption
  self.frames.captionLb:SetText( self.caption )
end

function GGF.Unit:SetAllianceRank( rank )
  if not self.template.RankLb or not self.template.RankTx then return end
  self.rank = rank
  if self.rank > 0 then
    GGF.Window:SetLabelText( self.frames.rankLb, self.rank, true )
    self.frames.rankTx:SetTexture( GetAvARankIcon(self.rank) )
    self.frames.rankLb:SetHidden( false )
    self.frames.rankTx:SetHidden( false )
  else
    self.frames.rankLb:SetHidden( true )
    self.frames.rankTx:SetHidden( true )
  end
end

function GGF.Unit:SetDifficulty( difficulty )
  if not self.template.Difficulty then return end
  self.difficulty = difficulty
  if self.difficulty == MONSTER_DIFFICULTY_NORMAL then
    self.frames.difficultyTx:SetTexture('ggFrames\\theme\\textures\\MonsterDiff1.dds')
    self.frames.difficultyTx:SetHidden(false)
  elseif self.difficulty == MONSTER_DIFFICULTY_HARD then
    self.frames.difficultyTx:SetTexture('ggFrames\\theme\\textures\\MonsterDiff2.dds')
    self.frames.difficultyTx:SetHidden(false)
  elseif self.difficulty == MONSTER_DIFFICULTY_DEADLY then
    self.frames.difficultyTx:SetTexture('ggFrames\\theme\\textures\\MonsterDiff3.dds')
    self.frames.difficultyTx:SetHidden(false)
  else
    self.frames.difficultyTx:SetHidden(true)
  end
end

function GGF.Unit:SetClass( classId )
  if not self.template.Class then return end
  self.class = classId
  if classId then
    self.frames.classTx:SetTexture( GGF.classTextures[classId] )
    self.frames.classTx:SetHidden(false)
  else
    self.frames.classTx:SetHidden(true)
  end
end

function GGF.Unit:SetExp( current, max, veteran )
  if not self.template.Experience or max == 0 then return end
  local exp = current / max
  if self.template.Experience.Label ~= false then
    self.frames.experienceLb:SetText( string.format("%d / %d", current, max) )
  end
  self.frames.experienceSt:SetWidth( exp * self.template.Experience.Bar.Width )
end

function GGF.Unit:SetLeader( isLeader )
  if not self.template.Leader then return end
  self.isLeader = isLeader
  self.frames.leaderTx:SetHidden( not isLeader )
  if self.template.Name ~= false and string.sub(self.unitName,0,10) ~= "LargeGroup" then
    GGF.Window:SetAdditionalLeftOffset( self.frames.nameLb, self.template.Name, isLeader and 20 or 0 )
  end
end

function GGF.Unit:SetStealth( isStealthed )
  if not self.template.Stealth then return end
  self.isStealthed = isStealthed
  self.frames.stealthTx:SetAlpha( isStealthed and 1 or 0.2 )
end

function GGF.Unit:SetDeath( isDead )
  self.isDead = isDead
  self:UpdateStatus()
  self.frames.death:SetHidden( not self.isDead )
end

function GGF.Unit:SetConnected( isOnline )
  if self.unitTag == 'player' then isOnline = true end
  self.isOnline = isOnline
  self:UpdateStatus()
  self.frames.offline:SetHidden( self.isOnline )
  if self.template.Class then self.frames.classTx:SetHidden( not self.isOnline ) end
  -- if isOnline and (not self.health or self.health.current == 0 ) then self:Load(self.unitTag) end  -- If the player was logged out the entire time we've been online, then his data was never truly loaded, lets reload
end

function GGF.Unit:UpdateStatus()
  local isDeadOrOffline = ( self.isDead or not self.isOnline)
  self.frames.healthBd:SetHidden( isDeadOrOffline )
  if self.frames.magickaBd then self.frames.magickaBd:SetHidden( isDeadOrOffline ) end
  if self.frames.staminaBd then self.frames.staminaBd:SetHidden( isDeadOrOffline ) end
  if self.frames.experienceBd and self.template.Experience ~= false then self.frames.experienceBd:SetHidden( isDeadOrOffline ) end
  if self.frames.stealthTx then self.frames.stealthTx:SetHidden( isDeadOrOffline ) end
  if self.frames.leaderTx then self.frames.leaderTx:SetHidden( isDeadOrOffline or not self.isLeader ) end
end

function GGF.Unit:SetRange( isWithinRange )
  self.frames.main:SetAlpha( isWithinRange and 1 or 0.5 )
end

function GGF.Unit:UpdateBuffs()
  if not GGF.SavedVars['Display_Buffs'] then return end
  local buffCount = { [BUFF_EFFECT_TYPE_BUFF] = 0, [BUFF_EFFECT_TYPE_DEBUFF] = 0 }
  for i = 1, GetNumBuffs(self.unitTag) do
    local buffName, timeStarted, timeEnding, buffSlot, stackCount, iconFilename, buffType, effectType, abilityType, statusEffectType = GetUnitBuffInfo(self.unitTag, i)
    local pool = self:GetBuffPool(effectType)
    if pool ~= nil then
      local buff = pool:AcquireObject(buffCount[effectType] + 1)  
      buff.timeStarted = timeStarted
      buff.timeEnding = timeEnding
      buff.neverEnding = (timeEnding == timeStarted)
      buff.data = { tooltipText = buffName }
      if (iconFilename == "/esoui/art/icons/icon_potion_full.dds") then
        iconFilename = "/esoui/art/icons/crafting_potion_base_water_2_r2.dds"
      end
      buff.icon:SetTexture(iconFilename)
      buff.text:SetText(buff.neverEnding and "" or buff.SecondsToString(timeEnding-timeStarted))
      buff:SetHidden(false)
      buffCount[effectType] = buffCount[effectType] + 1
    end
  end
  for eff, num in pairs(buffCount) do
    local pool = self:GetBuffPool(eff)
    for i = num+1, pool:GetActiveObjectCount() do
      pool:ReleaseObject(i)
    end
  end
end

function GGF.Unit:GetBuffPool(effectType)
  if effectType == BUFF_EFFECT_TYPE_BUFF then
    return self.buffPool
  elseif effectType == BUFF_EFFECT_TYPE_DEBUFF then
    return self.debuffPool
  else
    return nil
  end
end

function GGF.Unit:FormatStatusBarLabel( format, field )
  local str = tostring(format):gsub("Percentage", tostring(math.floor(self[field].percent)) )
                             :gsub("Max", tostring(self[field].max))
                             :gsub("Current", tostring(self[field].current))
                             :gsub("Nothing", "")

  if field == "health" and self.shield ~= nil and self.shield.current > 0 then
    str = str:gsub("+ Shield", tostring("+ "..self.shield.current))
  else
    str = str:gsub("+ Shield", "")
  end
  return str
end
